"""
greedy - Greedy (topological) coloring for GeoPandas

Copyright (C) 2019 Martin Fleischmann, 2017 Nyall Dawson

"""

import operator

__all__ = ["greedy"]


def _balanced(features, sw, balance="count", min_colors=4):
    """
    Strategy to color features in a way which is visually balanced.

    Algorithm ported from QGIS to be used with GeoDataFrames
    and libpysal weights objects.

    Original algorithm:
    Date                 : February 2017
    Copyright            : (C) 2017 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com

    Parameters
    ----------

    features : geopandas.GeoDataFrame
        GeoDataFrame.
    sw : libpysal.weights.W
        Spatial weights object denoting adjacency of features.
    balance : str (default 'count')
        The method of color balancing.
    min_colors : int (default 4)
        The minimal number of colors to be used.

    Returns
    -------

    feature_colors : dict
        Dictionary with assigned color codes.

    """
    feature_colors = {}
    # start with minimum number of colors in pool
    color_pool = set(range(min_colors))

    # calculate count of neighbours
    neighbour_count = sw.cardinalities

    # sort features by neighbour count - handle those with more neighbours first
    sorted_by_count = [
        feature_id
        for feature_id in sorted(
            neighbour_count.items(), key=operator.itemgetter(1), reverse=True
        )
    ]
    # counts for each color already assigned
    color_counts = {}
    color_areas = {}
    for c in color_pool:
        color_counts[c] = 0
        color_areas[c] = 0

    if balance == "centroid":
        features = features.copy()
        features.geometry = features.geometry.centroid
        balance = "distance"

    for (feature_id, n) in sorted_by_count:
        # first work out which already assigned colors are adjacent to this feature
        adjacent_colors = set()
        for neighbour in sw.neighbors[feature_id]:
            if neighbour in feature_colors:
                adjacent_colors.add(feature_colors[neighbour])

        # from the existing colors, work out which are available (ie non-adjacent)
        available_colors = color_pool.difference(adjacent_colors)

        feature_color = -1
        if len(available_colors) == 0:
            # no existing colors available for this feature; add new color and repeat
            min_colors += 1
            return _balanced(features, sw, balance, min_colors)
        else:
            if balance == "count":
                # choose least used available color
                counts = [
                    (c, v) for c, v in color_counts.items() if c in available_colors
                ]
                feature_color = sorted(counts, key=operator.itemgetter(1))[0][0]
                color_counts[feature_color] += 1
            elif balance == "area":
                areas = [
                    (c, v) for c, v in color_areas.items() if c in available_colors
                ]
                feature_color = sorted(areas, key=operator.itemgetter(1))[0][0]
                color_areas[feature_color] += features.loc[feature_id].geometry.area

            elif balance == "distance":
                min_distances = {c: float("inf") for c in available_colors}
                this_feature = features.loc[feature_id].geometry

                # find features for all available colors
                other_features = {
                    f_id: c
                    for (f_id, c) in feature_colors.items()
                    if c in available_colors
                }

                distances = features.loc[other_features.keys()].distance(this_feature)
                # calculate the min distance from this feature to the nearest
                # feature with each assigned color
                for other_feature_id, c in other_features.items():

                    distance = distances.loc[other_feature_id]
                    if distance < min_distances[c]:
                        min_distances[c] = distance

                # choose color such that min distance is maximised!
                # - ie we want MAXIMAL separation between features with the same color
                feature_color = sorted(
                    min_distances, key=min_distances.__getitem__, reverse=True
                )[0]

        feature_colors[feature_id] = feature_color

    return feature_colors


def greedy(
    gdf,
    strategy="balanced",
    balance="count",
    min_colors=4,
    sw="queen",
    min_distance=None,
    silence_warnings=True,
    interchange=False,
):
    """
    Color GeoDataFrame using various strategies of greedy (topological) colouring.

    Attempts to color a GeoDataFrame using as few colors as possible, where no
    neighbours can have same color as the feature itself. Offers various strategies
    ported from QGIS or implemented within NetworkX for greedy graph coloring.

    ``greedy`` will return ``pandas.Series`` representing assigned color codes.

    Parameters
    ----------

    gdf : GeoDataFrame
        GeoDataFrame
    strategy : str (default 'balanced')
        Determine coloring strategy. Options are ``'balanced'`` for
        algorithm based on QGIS Topological coloring. It is aiming
        for a visual balance, defined by the balance parameter. Other
        options are those supported by ``networkx.greedy_color``:

        * ``'largest_first'``
        * ``'random_sequential'``
        * ``'smallest_last'``
        * ``'independent_set'``
        * ``'connected_sequential_bfs'``
        * ``'connected_sequential_dfs'``
        * ``'connected_sequential'`` (alias for the previous strategy)
        * ``'saturation_largest_first'``
        * ``'DSATUR'`` (alias for the previous strategy)

        For details see https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.coloring.greedy_color.html

    balance : str (default 'count')
        If strategy is ``'balanced'``, determine the method of color balancing.

        * ``'count'`` attempts to balance the number of features per each color.
        * ``'area'`` attempts to balance the area covered by each color.
        * ``'centroid'`` attempts to balance the distance between colors based
          on the distance between centroids.
        * ``'distance'`` attempts to balance the distance between colors based
          on the distance between geometries. Slower than ``'centroid'``,
          but more precise.

        Both ``'centroid'`` and ``'distance'`` are significantly slower than other
        especially for larger GeoDataFrames. Apart from ``'count'``, all require
        CRS to be projected (not in degrees) to ensure metric values are correct.
    min_colors: int (default 4)
        If strategy is ``'balanced'``, define the minimal number of colors to be used.
    sw : 'queen', 'rook' or libpysal.weights.W (default 'queen')
        If min_distance is None, one can pass ``'libpysal.weights.W'``
        object denoting neighbors or let greedy generate one based on
        ``'queen'`` or ``'rook'`` contiguity.
    min_distance : float (default None)
        Set minimal distance between colors. If ``min_distance`` is not ``None``,
        slower algorithm for generating spatial weghts is used based on
        intersection between geometries. ``'min_distance'`` is then used as a
        tolerance of intersection.
    silence_warnings : bool (default True)
        Silence libpysal warnings when creating spatial weights.
    interchange : bool (default False)
        Use the color interchange algorithm (applicable for NetworkX strategies).
        For details see https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.coloring.greedy_color.html

    Returns
    -------

    color : pandas.Series
        ``pandas.Series`` representing assinged color codes.

    Examples
    --------

    >>> from mapclassify import greedy
    >>> import geopandas
    >>> world = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    >>> africa = world.loc[world.continent == "Africa"].copy()
    >>> africa = africa.to_crs("ESRI:102022").reset_index(drop=True)

    Default:

    >>> africa["greedy_colors"] = greedy(africa)
    >>> africa["greedy_colors"].head()
    0    1
    1    0
    2    0
    3    1
    4    4
    Name: greedy_colors, dtype: int64

    Balanced by area:

    >>> africa["balanced_area"] = greedy(africa, strategy="balanced", balance="area")
    >>> africa["balanced_area"].head()
    0    1
    1    2
    2    0
    3    1
    4    3
    Name: balanced_area, dtype: int64

    Using rook adjacency:

    >>> africa["rook_adjacency"] = greedy(africa, sw="rook")
    >>> africa["rook_adjacency"].tail()
    46    3
    47    0
    48    2
    49    3
    50    1
    Name: rook_adjacency, dtype: int64

    Adding minimal distance between colors:

    >>> africa["min_distance"] = greedy(africa, min_distance=1000000)
    >>> africa["min_distance"].head()
    0    1
    1    9
    2    0
    3    7
    4    4
    Name: min_distance, dtype: int64

    Using different coloring strategy:

    >>> africa["smallest_last"] = greedy(africa, strategy="smallest_last")
    >>> africa["smallest_last"].head()
    0    3
    1    1
    2    1
    3    3
    4    1
    Name: smallest_last, dtype: int64

    """  # noqa

    if strategy != "balanced":
        try:
            import networkx as nx

            STRATEGIES = nx.algorithms.coloring.greedy_coloring.STRATEGIES.keys()

        except ImportError:
            raise ImportError("The 'networkx' package is required.")

    try:
        import pandas as pd
    except ImportError:
        raise ImportError("The 'pandas' package is required.")
    try:
        from libpysal.weights import Queen, Rook, W, fuzzy_contiguity
    except ImportError:
        raise ImportError("The 'libpysal' package is required.")

    if min_distance is not None:
        sw = fuzzy_contiguity(
            gdf,
            tolerance=0.0,
            buffering=True,
            buffer=min_distance / 2.0,
            silence_warnings=silence_warnings,
        )

    if not isinstance(sw, W):
        if sw == "queen":
            sw = Queen.from_dataframe(gdf, silence_warnings=silence_warnings)
        elif sw == "rook":
            sw = Rook.from_dataframe(gdf, silence_warnings=silence_warnings)

    if strategy == "balanced":
        color = pd.Series(_balanced(gdf, sw, balance=balance, min_colors=min_colors))

    elif strategy in STRATEGIES:
        color = nx.greedy_color(
            sw.to_networkx(), strategy=strategy, interchange=interchange
        )

    else:
        raise ValueError(f"'{strategy}' is not a valid strategy.")

    color = pd.Series(color).sort_index()
    color.index = gdf.index
    return color



import numpy

from .classifiers import (
    BoxPlot,
    EqualInterval,
    FisherJenks,
    FisherJenksSampled,
    MaximumBreaks,
    Quantiles,
    StdMean,
    UserDefined,
)

__all__ = ["Pooled"]

dispatcher = {
    "boxplot": BoxPlot,
    "equalinterval": EqualInterval,
    "fisherjenks": FisherJenks,
    "fisherjenkssampled": FisherJenksSampled,
    "quantiles": Quantiles,
    "maximumbreaks": MaximumBreaks,
    "stdmean": StdMean,
    "userdefined": UserDefined,
}


class Pooled(object):
    """Applying global binning across columns.

    Parameters
    ----------

    Y : numpy.array
        :math:`(n, m)`, values to classify, with :math:`m>1`.
    classifier : str (default 'Quantiles')
        Name of ``mapclassify.classifier`` to apply.
    **kwargs : dict
        Additional keyword arguments for classifier.

    Attributes
    ----------

    global_classifier : mapclassify.classifiers.MapClassifier
        Instance of the pooled classifier defined as the classifier
        applied to the union of the columns.
    col_classifier : list
        Elements are ``MapClassifier`` instances with the pooled classifier
        applied to the associated column of ``Y``.

    Examples
    --------

    >>> import mapclassify
    >>> import numpy
    >>> n = 20
    >>> data = numpy.array([numpy.arange(n)+i*n for i in range(1,4)]).T
    >>> res = mapclassify.Pooled(data)

    >>> list(res.col_classifiers[0].counts)
    [12, 8, 0, 0, 0]

    >>> list(res.col_classifiers[1].counts)
    [0, 4, 12, 4, 0]

    >>> list(res.col_classifiers[2].counts)
    [0, 0, 0, 8, 12]

    >>> list(res.global_classifier.counts)
    [12, 12, 12, 12, 12]

    >>> res.global_classifier.bins == res.col_classifiers[0].bins
    array([ True,  True,  True,  True,  True])

    >>> res.global_classifier.bins
    array([31.8, 43.6, 55.4, 67.2, 79. ])

    """

    def __init__(self, Y, classifier="Quantiles", **kwargs):
        method = classifier.lower()
        valid_methods = list(dispatcher.keys())
        if method not in valid_methods:
            raise ValueError(
                f"'{classifier}' not a valid classifier. "
                f"Currently supported classifiers: {valid_methods}"
            )

        self.__dict__.update(kwargs)
        Y = numpy.asarray(Y)
        n, cols = Y.shape
        y = numpy.reshape(Y, (-1, 1), order="f")
        ymin = y.min()
        global_classifier = dispatcher[method](y, **kwargs)
        # self.k = global_classifier.k
        col_classifiers = []
        name = f"Pooled {classifier}"
        for c in range(cols):
            res = UserDefined(Y[:, c], bins=global_classifier.bins, lowest=ymin)
            res.name = name
            col_classifiers.append(res)
        self.col_classifiers = col_classifiers
        self.global_classifier = global_classifier
        self._summary()

    def _summary(self):
        self.classes = self.global_classifier.classes
        self.tss = self.global_classifier.tss
        self.adcm = self.global_classifier.adcm
        self.gadf = self.global_classifier.gadf

    def __str__(self):
        s = "Pooled Classifier"
        rows = [s]
        for c in self.col_classifiers:
            rows.append(c.table())
        return "\n\n".join(rows)

    def __repr__(self):
        return self.__str__()



import geopandas
import libpysal
import pytest

import mapclassify


def _assertions(a, b):
    assert a.k == b.k
    assert a.yb.all() == b.yb.all()
    assert a.bins.all() == b.bins.all()
    assert a.counts.all() == b.counts.all()


class TestClassify:
    def setup_method(self):
        link_to_data = libpysal.examples.get_path("columbus.shp")
        gdf = geopandas.read_file(link_to_data)
        self.x = gdf["HOVAL"].values

    def test_box_plot(self):
        a = mapclassify.classify(self.x, "box_plot")
        b = mapclassify.BoxPlot(self.x)
        _assertions(a, b)

    def test_equal_interval(self):
        a = mapclassify.classify(self.x, "EqualInterval", k=3)
        b = mapclassify.EqualInterval(self.x, k=3)
        _assertions(a, b)

    def test_fisher_jenks(self):
        a = mapclassify.classify(self.x, "FisherJenks", k=3)
        b = mapclassify.FisherJenks(self.x, k=3)
        _assertions(a, b)

    def test_fisher_jenks_sampled(self):
        a = mapclassify.classify(
            self.x, "FisherJenksSampled", k=3, pct_sampled=0.5, truncate=False
        )
        b = mapclassify.FisherJenksSampled(self.x, k=3, pct=0.5, truncate=False)
        _assertions(a, b)

    def test_headtail_breaks(self):
        a = mapclassify.classify(self.x, "headtail_breaks")
        b = mapclassify.HeadTailBreaks(self.x)
        _assertions(a, b)

    def test_quantiles(self):
        a = mapclassify.classify(self.x, "quantiles", k=3)
        b = mapclassify.Quantiles(self.x, k=3)
        _assertions(a, b)

    def test_percentiles(self):
        a = mapclassify.classify(self.x, "percentiles", pct=[25, 50, 75, 100])
        b = mapclassify.Percentiles(self.x, pct=[25, 50, 75, 100])
        _assertions(a, b)

    def test_jenks_caspall(self):
        a = mapclassify.classify(self.x, "JenksCaspall", k=3)
        b = mapclassify.JenksCaspall(self.x, k=3)
        _assertions(a, b)

    def test_jenks_caspall_forced(self):
        a = mapclassify.classify(self.x, "JenksCaspallForced", k=3)
        b = mapclassify.JenksCaspallForced(self.x, k=3)
        _assertions(a, b)

    def test_jenks_caspall_sampled(self):
        a = mapclassify.classify(self.x, "JenksCaspallSampled", pct_sampled=0.5)
        b = mapclassify.JenksCaspallSampled(self.x, pct=0.5)
        _assertions(a, b)

    def test_natural_breaks(self):
        a = mapclassify.classify(self.x, "natural_breaks")
        b = mapclassify.NaturalBreaks(self.x)
        _assertions(a, b)

    def test_max_p_classifier(self):
        a = mapclassify.classify(self.x, "max_p", k=3, initial=50)
        b = mapclassify.MaxP(self.x, k=3, initial=50)
        _assertions(a, b)

    def test_std_mean(self):
        a = mapclassify.classify(self.x, "std_mean", multiples=[-1, -0.5, 0.5, 1])
        b = mapclassify.StdMean(self.x, multiples=[-1, -0.5, 0.5, 1])
        _assertions(a, b)

    def test_user_defined(self):
        a = mapclassify.classify(self.x, "user_defined", bins=[20, max(self.x)])
        b = mapclassify.UserDefined(self.x, bins=[20, max(self.x)])
        _assertions(a, b)

    def test_bad_classifier(self):
        classifier = "George_Costanza"
        with pytest.raises(ValueError, match="Invalid scheme: 'georgecostanza'"):
            mapclassify.classify(self.x, classifier)



import sys

import geopandas
import libpysal
import pytest

from ..greedy import greedy

PY39 = sys.version_info.major == 3 and sys.version_info.minor == 9

world = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
sw = libpysal.weights.Queen.from_dataframe(
    world, ids=world.index.to_list(), silence_warnings=True
)


@pytest.mark.filterwarnings("ignore:Geometry is in a geographic CRS.")
class TestGreedy:
    def test_default(self):
        colors = greedy(world)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [36, 36, 35, 35, 35]
        assert (colors.index == world.index).all()

    def test_rook(self):
        colors = greedy(world, sw="rook")
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [36, 36, 35, 35, 35]

    def test_sw(self):
        colors = greedy(world, sw=sw)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [36, 36, 35, 35, 35]

    def test_min_distance(self):
        europe = world.loc[world.continent == "Europe"].to_crs(epsg=3035)
        colors = greedy(europe, min_distance=500000)
        assert len(colors) == len(europe)
        assert set(colors) == set(range(13))
        assert colors.value_counts().to_list() == [3] * 13

    def test_invalid_strategy(self):
        strategy = "spice melange"
        with pytest.raises(ValueError, match=f"'{strategy}' is not a valid strategy."):
            greedy(world, strategy=strategy)


@pytest.mark.filterwarnings("ignore:Geometry is in a geographic CRS.")
@pytest.mark.parametrize("pysal_geos", [None, 0])
class TestGreedyParams:
    def test_count(self, pysal_geos):
        colors = greedy(
            world, strategy="balanced", balance="count", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [36, 36, 35, 35, 35]

    def test_area(self, pysal_geos):
        colors = greedy(
            world, strategy="balanced", balance="area", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [55, 49, 39, 32, 2]

    def test_centroid(self, pysal_geos):
        colors = greedy(
            world, strategy="balanced", balance="centroid", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [39, 36, 36, 34, 32]

    def test_distance(self, pysal_geos):
        colors = greedy(
            world, strategy="balanced", balance="distance", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [38, 36, 35, 34, 34]

    def test_largest_first(self, pysal_geos):
        colors = greedy(world, strategy="largest_first", min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [64, 49, 42, 21, 1]

    def test_random_sequential(self, pysal_geos):
        """based on random, no consistent results to be tested"""
        colors = greedy(world, strategy="random_sequential", min_distance=pysal_geos)
        assert len(colors) == len(world)

    def test_smallest_last(self, pysal_geos):
        colors = greedy(world, strategy="smallest_last", min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3])
        # skip pn Python 3.9 due to networkx/networkx#3993
        if not PY39:
            assert colors.value_counts().to_list() == [71, 52, 39, 15]

    def test_independent_set(self, pysal_geos):
        colors = greedy(world, strategy="independent_set", min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [91, 42, 26, 13, 5]

    def test_connected_sequential_bfs(self, pysal_geos):
        colors = greedy(
            world, strategy="connected_sequential_bfs", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [77, 46, 34, 18, 2]

    def test_connected_sequential_dfs(self, pysal_geos):
        colors = greedy(
            world, strategy="connected_sequential_dfs", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [75, 52, 34, 14, 2]

    def test_connected_sequential(self, pysal_geos):
        colors = greedy(world, strategy="connected_sequential", min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [77, 46, 34, 18, 2]

    def test_saturation_largest_first(self, pysal_geos):
        colors = greedy(
            world, strategy="saturation_largest_first", min_distance=pysal_geos
        )
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3])
        assert colors.value_counts().to_list() == [71, 47, 42, 17]

    def test_DSATUR(self, pysal_geos):
        colors = greedy(world, strategy="DSATUR", min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3])
        assert colors.value_counts().to_list() == [71, 47, 42, 17]

    def test_index(self, pysal_geos):
        world["ten"] = world.index * 10
        reindexed = world.set_index("ten")
        colors = greedy(reindexed, min_distance=pysal_geos)
        assert len(colors) == len(world)
        assert set(colors) == set([0, 1, 2, 3, 4])
        assert colors.value_counts().to_list() == [36, 36, 35, 35, 35]
